// 2011.09.08: Header is included from util.h.
//#include <stdio.h>
//#include <math.h>
//#include <stdlib.h>
#include <string.h>
//#include "mkl_lapack.h"
//#include "mkl_cblas.h"

#include "util.h"


void printUsage(){
	printf("\n");
	printf("Usage: call [options] -n N -i INPUT -g GMM -o OUTPUT [> logfile]\n");
	printf("\n");
	printf("Call individual's genotypes using the esitimated GMM for each CNP locus.\n");
	printf("\n");
	printf("Options:\n");
	printf("  -h, --help        Print short help message and exit\n");
	printf("  --version         Print version info and exit\n");
	printf("  -i                INPUT file name of signal intensity data (binary, double-precision)\n");
	printf("  -n                Sample size N\n");
	printf("  -s                Subsamples specified by a series of 0/1 values (binary, integer)\n");
	printf("  -g                File name of estimated parameters and QC statistics obtained via GMM fitting \n");
	printf("                    (Only the binary output file of gmm. The text file is not available.)\n");
	printf("\n");
	printf("  -o                Prefix of OUTPUT file(s)\n");
	printf("                    organized as CNP (row) by individuals (column). \n");
	printf("  -a                Output MAP allele-specific CN genotypes [0..14] as OUTPUT.ascn (binary, integer)\n");
	printf("\n");
	printf("                    Cluster assignment:  \n");
	printf("                    Genotype        Code  \n");
	printf("                    --------------------  \n");
	printf("                    O                  0  \n");
	printf("                    A                  1  \n");
	printf("                    AA                 2  \n");
	printf("                    AAA                3  \n");
	printf("                    AAAA               4  \n");
	printf("                    B                  5  \n");
	printf("                    AB                 6  \n");
	printf("                    AAB                7  \n");
	printf("                    AAAB               8  \n");
	printf("                    BB                 9  \n");
	printf("                    ABB               10  \n");
	printf("                    AABB              11  \n");
	printf("                    BBB               12  \n");
	printf("                    ABBB              13  \n");
	printf("                    BBBB              14  \n");
	printf("                    Undetermined      99  \n");
	printf("                    --------------------  \n");
	printf(" \n");
	printf("  -c                Output MAP CN dosages [0..4] as OUTPUT.map.cn (binary, integer; Undetermined=99)\n");
	printf("  -d                Output posterior mean CN dosages [0,4] as OUTPUT.mean.cn\n");
	printf("                    (binary, double-precision; Undetermined=NaN)\n");
	printf("  -l                Output MAP allelic variations [-4..4] as OUTPUT.map.av (binary integer; Undetermined=99)\n");
	printf("  -m                Output posterior mean allelic variations [-4,4] as OUTPUT.mean.av\n");
	printf("                    (binary double-precision; Undetermined=NaN)\n");
	printf("  --cr              Call rate threshold (0.99 as default)\n");
	printf("  --hwep            P-value of Hardy-Weinberg test statistic threshold (1e-6 as default)\n");
	printf("  --err             The estimated error rate threshold (0.1 as default)\n");
	printf("  --mahf            The minimum aneuploid haplotype frequency threshold (0.01 as defult)\n");
	printf("  -f, --force       Call genotypes for all loci without QC\n");
	printf("  -t, --text        Output as text\n");
	printf("  -v, --verbose     Print more information about progress\n");
	//printf("  -q, --quiet           Don't print any message\n");
	printf("\n");
	printf("Report bugs to <kumasaka@src.riken.jp>.\n\n");
}



int main(int argc, char **argv){
	int itr,i,j,k,flag;
	int inputfileid=0;
	int outputfileid=0;
	int gmmfileid=0;
	double tcr, thwep, terr, tmahf;
	
	FILE *fp, *op,  *fsubset;//*fclus, *fdose, ;
	FILE *fascn, *fcn, *fav, *fmcn, *fmav, *fstat;
	int qcstatslen = 6;
	char *postfix[6] = {".ascn",".map.cn",".mean.cn",".map.av",".mean.av",".status"};
	char* tmpfilename;
	int verbose, text, force;
	verbose=0;
	force=0;
	text=0;
	tcr = 0.99;
	thwep = 1.0e-6;
	terr = 0.1;
	tmahf = 0.01;
	
	if(argc<2){printUsage(); return 0;}
	for(i=0; i<argc; i++){if(strcmp(argv[i],"-h")==0 || strcmp(argv[i],"--help")==0){printUsage(); return 0;}}
	for(i=0; i<argc; i++){if(strcmp(argv[i],"-v")==0 || strcmp(argv[i],"--verbose")==0){verbose=1;}}
	for(i=0; i<argc; i++){if(strcmp(argv[i],"-f")==0 || strcmp(argv[i],"--force")==0){force=1;}}
	for(i=0; i<argc; i++){if(strcmp(argv[i],"-t")==0 || strcmp(argv[i],"--text")==0){text=1;}}
	for(i=0; i<argc; i++){if(strcmp(argv[i],"--version")==0){printVersion(); return 0;}}
	
	
	printVersion();
	printf("User command:\n");
	for(i=0; i<argc; i++){printf("%s ", argv[i]);}
	printf("\n\n");
	flag=0;
	for(i=0; i<argc; i++){if(strcmp(argv[i],"-i")==0 && argc>i+1){if((fp=fopen(argv[i+1],"rb"))==NULL){printf("No intensity file!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);}else{printf("Intensity file: %s\n", argv[i+1]);flag=1;inputfileid = i+1;};break;}}
	if(flag==0){printf("No intensity file!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);}
	
	flag=0;
	for(i=0; i<argc; i++){if(strcmp(argv[i],"-g")==0 && argc>i+1){if((op=fopen(argv[i+1],"rb"))==NULL){printf("No GMM file!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);}else{printf("GMM file:       %s\n", argv[i+1]);flag=1;};  gmmfileid = i+1; break;}}
	if(flag==0){printf("No GMM file!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);}
	
	flag=0;
	for(i=0; i<argc; i++){if(strcmp(argv[i],"-s")==0 && argc>i+1){if((fsubset = fopen(argv[i+1],"rb"))==NULL){printf("No subset file exists!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);}else{printf("Subset file:    %s\n", argv[i+1]);flag=1;};break;}}
	if(flag==0){printf("Subset file: no\n"); fsubset=NULL;}
	
	flag=0;
	for(i=0; i<argc; i++){if(strcmp(argv[i],"-o")==0 && argc>i+1){if(argv[i+1][0]!='-'){flag=1; outputfileid = i+1; break;}}}
	if(flag==0){printf("No output prefix specified!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);}
	tmpfilename = (char*)calloc(strlen(argv[outputfileid])+10,sizeof(char));
	sscanf(argv[outputfileid], "%s", tmpfilename);
	
	
	printf("\n");
	flag=0;
	for(i=0; i<argc; i++){
		if(strcmp(argv[i],"-a")==0){
			sscanf(postfix[0], "%s", tmpfilename+strlen(argv[outputfileid]));
			if((fascn = fopen(tmpfilename,"wb"))==NULL){
				printf("No ASCN genotype file available!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);
			}else{
				printf("ASCN genotypes:          %s\n", tmpfilename);flag=1;
			}
			break;
		}
	}
	if(flag==0){printf("ASCN genotypes:          no\n"); fascn=NULL;}
	
	flag=0;
	for(i=0; i<argc; i++){
		if(strcmp(argv[i],"-c")==0){
			sscanf(postfix[1], "%s", tmpfilename+strlen(argv[outputfileid]));
			if((fcn = fopen(tmpfilename,"wb"))==NULL){
				printf("No MAP CN dosage file available!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);
			}else{
				printf("MAP CN dosages:          %s\n", tmpfilename);flag=1;
			}
			break;
		}
	}
	if(flag==0){printf("MAP CN dosages:          no\n"); fcn=NULL;}
	
	flag=0;
	for(i=0; i<argc; i++){
		if(strcmp(argv[i],"-d")==0){
			sscanf(postfix[2], "%s", tmpfilename+strlen(argv[outputfileid]));
			if((fmcn = fopen(tmpfilename,"wb"))==NULL){
				printf("No mean CN dosage file available!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);
			}else{
				printf("Mean CN dosages:         %s\n", tmpfilename);flag=1;
			}
			break;
		}
	}
	if(flag==0){printf("Mean CN dosages:         no\n"); fmcn=NULL;}
	
	
	
	flag=0;
	for(i=0; i<argc; i++){
		if(strcmp(argv[i],"-l")==0){
			sscanf(postfix[3], "%s", tmpfilename+strlen(argv[outputfileid]));
			if((fav = fopen(tmpfilename,"wb"))==NULL){
				printf("No MAP allelic variation file available!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);
			}else{
				printf("MAP allelic variations:  %s\n", tmpfilename);flag=1;
			}
			break;
		}
	}
	if(flag==0){printf("MAP allelic variations:  no\n"); fav=NULL;}
	
	
	flag=0;
	for(i=0; i<argc; i++){
		if(strcmp(argv[i],"-m")==0){
			sscanf(postfix[4], "%s", tmpfilename+strlen(argv[outputfileid]));
			if((fmav = fopen(tmpfilename,"wb"))==NULL){
				printf("No mean allelic variation file available!\n\nPlatinumCNV terminated...\n\n");
				exit(EXIT_FAILURE);
			}else{
				printf("Mean allelic variations: %s\n", tmpfilename);flag=1;
			}
			break;
		}
	}
	if(flag==0){printf("Mean allelic variations: no\n"); fmav=NULL;}
	if(fascn==NULL&&fav==NULL&&fcn==NULL&&fmcn==NULL&&fmav==NULL){
		printf("\nNon of output file types -[acdlm] is specified!\n\nPlatinumCNV terminated...\n\n");
		exit(EXIT_FAILURE);
	}
	
	//genotyping call status
	sscanf(postfix[5], "%s", tmpfilename+strlen(argv[outputfileid]));
	if((fstat = fopen(tmpfilename,"wb"))==NULL){
		printf("No call status file can be opened!\n\nPlatinumCNV terminated...\n\n");
		exit(EXIT_FAILURE);
	}else{
		printf("Genotype call status:    %s\n", tmpfilename);
		fprintf(fstat,"ID\tisCalled\tCallRate\tLog10HWEP\tErrorRate\n");
	}
	
	
	
	
	
	printf("\n\n");
	
	
	
	flag=0;
	for(i=0; i<argc; i++){if(strcmp(argv[i],"-n")==0 && argc>i+1){if(sscanf(argv[i+1], "%d", &N)==0){printf("Sample size is invalid!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);}else{printf("Sample size: %d\n", N);}; flag=1;}}
	if(flag==0){printf("Sample size is missing!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);}
	if(N<15){printf("Sample size is too small!\n\nPlatinumCNV terminated...\n\n");exit(EXIT_FAILURE);}
	
	int nl=0;
	ldiv_t nlt = ncol(argv[inputfileid], 2*N);
	if(nlt.rem==0){
		nl = (int)nlt.quot;
		printf("Marker size: %d\n", nl);
	}else{
		printf("Sample size is inconsistent with the intensity file size! (size of %s = 8 * %ld * %d + %ld) \n\nPlatinumCNV terminated...\n\n",argv[inputfileid],nlt.quot,N*2,nlt.rem);exit(EXIT_FAILURE);
	}
	nlt = ncol(argv[gmmfileid], qcstatslen+K*7+J);
	if(nlt.rem!=0){
		printf("The column number of GMM file is inconsistent with %d! (size of %s = 8 * %ld * %d + %ld) \n\nPlatinumCNV terminated...\n\n", qcstatslen+K*7+J, argv[gmmfileid], nlt.quot, qcstatslen+K*7+J, nlt.rem);exit(EXIT_FAILURE);
	}else if(nlt.quot != nl){
		printf("The row number of GMM file is inconsistent with that of the intensity file! (size of %s = 8 * %ld * %d + %ld) \n\nPlatinumCNV terminated...\n\n", argv[gmmfileid], nlt.quot, qcstatslen+K*7+J, nlt.rem);exit(EXIT_FAILURE);
	}
	
	
	
	printf("\n\n");
	
	
	
	double lkhd;
	
	double *Theta;
	double *Theta15;
	double *eTheta15;

	double *Mu;
	double *Sigma;
	double *Mutemp;
	double *Sigmatemp;
	double *xt;
	
	double* dose_cn;
	double* dose_av;
	int* clus;
	int* clus_av;
	int* clus_cn;
	
	double* Z6;
	double* Z15;
	double* Z21;
	double* d;
	double* x0; 
	double* x;
	double* w;
	int* wtmp;
	int* tab;
	int* subset;

	int N0, Nw;
	
	double cr[2];
	double err=0.0;
	double qcstats[qcstatslen];
	
	
	Theta = (double*)calloc(J, sizeof(double));
	Theta15 = (double*)calloc(K, sizeof(double));
	eTheta15 = (double*)calloc(K, sizeof(double));
	
	Mu = (double*)calloc(K*2, sizeof(double));
	Sigma = (double*)calloc(4*K, sizeof(double));
	Mutemp = (double*)calloc(K*2, sizeof(double));
	Sigmatemp = (double*)calloc(K*4, sizeof(double));
	xt = (double*)calloc(2, sizeof(double));
	
	
	hess = (double*)calloc(K*4*K, sizeof(double));
	
	subset = (int*)calloc(N, sizeof(int));
	x0 = (double*)calloc(N*2, sizeof(double));
	x  = (double*)calloc(N*2,  sizeof(double));
	w = (double*)calloc(N,  sizeof(double));
	wtmp = (int*)calloc(N,  sizeof(int));
	tab = (int*)calloc(N,  sizeof(int));
	
	Z6 = (double*)calloc(N*J, sizeof(double));
	Z15 = (double*)calloc(N*K, sizeof(double));
	d = (double*)calloc(N, sizeof(double));
	Z21 = (double*)calloc(N*(J+1)*J/2, sizeof(double));
	X1 = (double*)calloc(N*2, sizeof(double));
	X2 = (double*)calloc(N*2, sizeof(double));
	X3 = (double*)calloc(N*2, sizeof(double));	
	dose_av = (double*)calloc(N, sizeof(double));
	dose_cn = (double*)calloc(N, sizeof(double));
	clus = (int*)calloc(N, sizeof(int));
	clus_av = (int*)calloc(N, sizeof(int));
	clus_cn = (int*)calloc(N, sizeof(int));
	uvn = (double*)calloc(N, sizeof(double));
	lk = (double*)calloc(500, sizeof(double));
	uvk = (double*)calloc(K, sizeof(double));
	uvj = (double*)calloc(J*(J+1)/2, sizeof(double));
	
	for(i=0; i<N; i++){uvn[i] = 1.0;}
	for(i=0; i<K; i++){uvk[i] = 1.0;}
	for(i=0; i<J*(J+1)/2; i++){uvj[i] = 1.0;}
	
	if(X1 == NULL || X2 == NULL || X3 == NULL || Z6 == NULL || Z15 == NULL ||Z21 == NULL || uvn == NULL || x == NULL || x0 == NULL || w == NULL || wtmp == NULL || tab == NULL || subset == NULL){printf("Memory allocation failure...\nSample size (%d) is too large!\n", N); exit(EXIT_FAILURE);}else{printf("Memory allocated successfully...\n\n");}

	
	for(i=0; i<argc; i++){if(strcmp(argv[i],"--cr")==0 && argc>i+1)  {if(sscanf(argv[i+1], "%lf", &tcr)==0){  printf("Call rate is invalid...use the default value of 0.99.\n"); tcr=0.99;}; break;}}
	for(i=0; i<argc; i++){if(strcmp(argv[i],"--hwep")==0 && argc>i+1){if(sscanf(argv[i+1], "%lf", &thwep)==0){printf("HWE P-value is invalid...use the default value of 1e-6.\n"); thwep=1.0e-6;}; break;}}
	for(i=0; i<argc; i++){if(strcmp(argv[i],"--err")==0 && argc>i+1) {if(sscanf(argv[i+1], "%lf", &terr)==0){ printf("Error rate is invalid...use the default value of 0.1.\n"); terr=0.1;}; break;}}
	for(i=0; i<argc; i++){if(strcmp(argv[i],"--mahf")==0 && argc>i+1){if(sscanf(argv[i+1], "%lf", &tmahf)==0){printf("Call rate is invalid...use the default value of 0.01.\n"); tmahf=0.01;}; break;}}
	printf("\n\n");
	printf("Call rate threshold:                 %.3lf\n", tcr);
	printf("HWE P-value threshold:               %.2e\n", thwep);
	printf("Genotyping error rate threshold:     %.3lf\n", terr);
	printf("Aneuploid haplotype freq. threshold: %.3lf\n", tmahf);
	printf("\n\n");
	
	if(fsubset==NULL){
		//printf("odata not found!\n");exit(EXIT_FAILURE);
		for(i=0; i<N; i++){subset[i] = 1;}
		N0 = N;
	}else{ 
		//printf("subset=T\n");
		N0 = N;
		fread(subset, sizeof(int), N0, fsubset);
		N = 0;
		for(i=0; i<N0; i++){if(subset[i]==1){N++;}}
	}
	printf("N of subjects used: %d/%d\n\n\n", N, N0);
	
	
	alpha0 = cblas_dasum(J, alpha, 1);
	cblas_dcopy(K*2, Mu0, 1, Mutemp, 1);
	cblas_dcopy(K*4, Sigma0, 1, Sigmatemp, 1);
	
	
	
	for(itr=0; itr<nl; itr++){
		int called=0;
		printf("Marker %d\t",itr+1);
		fread(x0, sizeof(double), 2*N0, fp);
		j=0;
		
		fread(Mu, sizeof(double), 2*K, op);
		fread(Sigma, sizeof(double), 4*K, op);
		fread(Theta15, sizeof(double), K, op);
		fread(Theta, sizeof(double), J, op);
		fread(qcstats, sizeof(double), qcstatslen, op);

if(verbose==1){
vprint(Mu, K*2);vprint(Sigma, K*4);vprint(Theta, J);vprint(qcstats, qcstatslen);
}

		int npe=0.0;
		for(k=0; k<N0; k++){
			if(subset[k]==1){
				//xt[0] = 1.006711*pow(x0[2*k],0.9317417);
				//xt[1] = pow(x0[2*k+1],1.073259)/1.006711;
				xt[0] = x0[2*k];
				xt[1] = x0[2*k+1];
				if(isnan(xt[0])==0 && isnan(xt[1])==0){
					w[j]=1.0;
					npe++;
				}else{
					w[j]=0.0;
				}
				x[j*2] = xt[0];
				x[j*2+1] = xt[1];
				j++;
			}
		}
		Nw = j;
		
if(verbose==1)printf("N of subjects available=%d", Nw);
		
		if(qcstats[0]<5.0){
			
			Estep(x, w, Nw, Mu, Sigma, Theta, Z15, Z21, d, &lkhd);
			
			getETheta15(Theta,eTheta15);
			
			// Call Rate
			cr[0]=cr[1]=0.0;
			for(i=0; i<Nw; i++){ if(d[i]>0.0){cr[0]+=w[i];}; if(d[i]>(0.05/(double)N)){cr[1]+=w[i];} }
			// HWE Pvalue
			double pchisq=0.0;
			getChisq(Theta15, Theta, cr[0], &pchisq);
			// Error rate
			err = getErrorRate(Z15, Theta15, w, Nw);
			
			if(isnan(err)>0){err=1.0;}
			if(isnan(cr[0])>0 || isnan(cr[1])>0){cr[1]=0.0;cr[0]=1.0;}
			if(isnan(pchisq)>0){pchisq=.0;}
			
if(verbose==1)printf("Call rate: %.3lf HWE P-value: %.2e Error rate: %.3lf",cr[1]/cr[0], pchisq, err);
			
			if(force==1 || (cr[1]/cr[0]>tcr && pchisq > thwep && err < terr && chktheta(Theta, tmahf)==0 && qcstats[0]<0.5)){
				printf("Call");
				called=1;
				getClus(Z15,   Nw, clus);
				getClusCN(Z15, Nw, clus_cn);
				getClusAV(Z15, Nw, clus_av);
				getDose(Z15,   Nw, dose_cn);
				getDoseAV(Z15, Nw, dose_av);
				for(i=0; i<N; i++){
					if(d[i]<0.01/((double)Nw)){
						clus[i]=clus_cn[i]=clus_av[i]=99;
						dose_av[i]=dose_cn[i]=log(-1.0);
					}
				}
				if(text==1){
					FwriteTextInt(clus, Nw, fascn);
					FwriteTextInt(clus_cn, Nw, fcn);
					FwriteTextInt(clus_av, Nw, fav);
					FwriteText(dose_cn, Nw, fmcn);
					FwriteText(dose_av, Nw, fmav);
				}else{
					FwriteInt(clus, Nw, fascn);
					FwriteInt(clus_cn, Nw, fcn);
					FwriteInt(clus_av, Nw, fav);
					Fwrite(dose_cn, Nw, fmcn);
					Fwrite(dose_av, Nw, fmav);
				}
			}
			fprintf(fstat,"Marker %d\t%d\t%.3lf\t%.2lf\t%.3lf\n", itr+1, called, cr[1]/cr[0], log10(pchisq), err);
		}else{
			printf("No data points!\n");
			fprintf(fstat,"Marker %d\t0\tNA\tNA\tNA\n",itr+1);
			if(force==1){
				if(text==1){
					FwriteTextInt(NULL, Nw, fascn);
					FwriteTextInt(NULL, Nw, fcn);
					FwriteTextInt(NULL, Nw, fav);
					FwriteText(NULL, Nw, fmcn);
					FwriteText(NULL, Nw, fmav);
				}else{
					FwriteInt(NULL, Nw, fascn);
					FwriteInt(NULL, Nw, fcn);
					FwriteInt(NULL, Nw, fav);
					Fwrite(NULL, Nw, fmcn);
					Fwrite(NULL, Nw, fmav);
				}
			}
		}
		printf("\n");
	}
	free(Z15);
	free(Z21);
	free(uvk);
	free(uvn);
	free(uvj);
	
	//if(fdose!=NULL){fclose(fdose);}
	//fclose(fid);
	fclose(op);
	fclose(fp);
	//if(fclus!=NULL){fclose(fclus);}
	
	return 0;
}














