// 2011.09.08: Add include guarde, to avoid double inclusion.
#ifndef _UTIL_H
#define _UTIL_H

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

// 2011.09.08: To support atlas.
#ifdef USEMKL
#include "mkl_cblas.h"
#include "mkl_lapack.h"
#elif USEATLAS
#include <clapack.h>
#endif

int mvdnorm(double* X, int N, double* Mu, double* Sigma, double* z);
int inverse(double* mat, double* inv);
void GMM(double* X, double* w, int N, double* Mu, double* Sigma, double* Theta, double* Theta15, double* Theta21, double* Z15, double* Z21, double* d, double* lkhd, int* status);
void Mstep(double* X1, double* w, int N, double* Z15, double* Z21, double* Theta, double* Theta15, double* Theta21, double* Mu, double* Sigma);
void Estep(double* X, double* w, int N, double* Mu, double* Sigma, double* Theta, double* Z15, double* Z21, double* d, double* lk);
void clear(double* v, int N);
void clearInt(int* v, int N);
void trans(double* x, int n, double rho);
void table(int* x, int N, int* t, double* w, int* Nw);
void mprint(double* v, int n);
void vprint(double* v, int n);
void vprintInt(int* v, int n, int offset);
void viprint(int* v, int n);
double Power(double x, int n);
double det(double* mat);
double DKL(double* m0, double* m1, double* s0, double* s1);
double quad(double* mat, double* vec);
double EP(double* m0, double* m1, double* s0);
int diff(double* x, double* y, double* z);
int isna(double* v, int N);
void getZ6(double* Z6, double* Z21);
double getErrorRate(double* Z15, double* Theta15, double* w, int Nw);
void setAVar(double* X, double* Z21, double* Z15, double* Z6, double* Theta, double* Mu, double* Sigma, double* cjg, double* dig, double nep, double* avar);
void getChisq(double* Theta15orig, double* Theta6, double cr, double* chisq);
void getETheta15(double* Theta, double* eTheta15);
void getFreq(double* Z15, double* d, int* clus, double* freq, double th);
void getFreq0(double* ZAB, double* thetaAB, int N, int K, int* clus, double* freq, double th);
double max(double* v, int n);
double min(int n, double* v, int k);
double min0(double* v, int n);
double mdist(double m1, double m2, double s1, double s2);
int chklattice(double* Mu, double* Sigma, double* Theta15);
int chktheta(double* Theta, double th);
int chkABtheta(double* Theta);
int chkEP(double* Mu, double* Sigma, double* Theta15);
double dkls[225];
double getEP(int i, int j);
void normalize(double* x, int n);
void settheta(double* Theta, double q1, double q2);
int chol(double* A, double* L);
ldiv_t ncol(char* fn, int N);
int ncol_int(char* fn, int N);
double getMaha(double* Mu, double* Mu0);

void getClus(double* Z15, int N, int* clus);
void getClusCN(double* Z15, int N, int* clus);
void getClusAV(double* Z15, int N, int* clus);
void getDose(double* Z15, int N, double* dose);
void getDoseAV(double* Z15, int N, double* dose);


int cs[25];

double* X1;
double* X2;
double* X3;

double* uvn;
double* uvk;
double* uvj;

void printVersion();
void printUsage();


double* lk;

int K;
int J;
int N;

int a[6];
int b[6];
double ab[15];
double cn[15];
double m[30]; // 2*K log(Mu0) prior 

double alpha[6];
double alpha0;
double delta0;
double lambda0;

double ng[15];
double ngxb[30];
double Nu[30];
double Nu0[30];
double grad[30];
double* hess;

double Mu0[30];
double Sigma0[60];
double SigmaInv[60];
double Sinv[900];
double hessTemp[900];
double S[30];


double polygamma(int n, double x);
double lgamma(double x);
double digamma(double x);
double trigamma(double x); 
double q_chisq(double, int);

double ddirichletmultinom(double* z, double* a, int J);
void dmvdirichletmultinom(double* a, int J, double* p21);
double parLDM(double* a, double* p21, int J);


void FwriteTextInt(int* x, int N, FILE* fp);
void FwriteText(double* x, int N, FILE* fp);
void FwriteInt(int* x, int N, FILE* fp);
void Fwrite(double* x, int N, FILE* fp);

#endif /* _UTIL_H */
